##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Kyle A. Beauchamp
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import mdtraj as md
import numpy as np
import pandas as pd
import pytest
from mdtraj.testing import eq

temperature = 300.
tip3p_charges = np.array([-0.834, 0.417, 0.417])

"""
Reference value taken from Gromacs 5.0.4 using the following commands.
Note that gromacs accumulates property averages over every timestep, so precise
comparisons require writing XTC output very frequently.


mkdir gromacs
cd gromacs
ln -s ../MDTraj/testing/reference/ ./reference

pdb2gmx -ff amber99sb -water tip3p -f ./reference/tip3p_300K_1ATM.pdb  -ignh
grompp -f ./reference/md.mdp -c ./conf.gro -p ./topol.top

# Do a quick equilibration run
export OMP_NUM_THREADS=1
mdrun -nsteps 500000

# Do a short production run
grompp -f ./reference/md.mdp -c ./confout.gro -p ./topol.top
mdrun -nsteps 200000
cp traj_comp.xtc reference/tip3p_300K_1ATM.xtc

# Compute tab separated dataset for energies
g_energy -xvg None
15
cp energy.xvg reference/tip3p_300K_1ATM.csv
(save enthalpy to csv style output)

# Compute static dielectric constant
g_dipoles

Epsilon = 87.1818

# Compute fluctuation properties
g_energy -fluct_props

Energy                      Average   Err.Est.       RMSD  Tot-Drift
-------------------------------------------------------------------------------
Kinetic En.                 1916.68          3    66.4762    9.59656  (kJ/mol)
Total Energy               -8286.19        9.9    120.584    28.1681  (kJ/mol)
Temperature                 298.411       0.46    10.3498     1.4941  (K)
Pressure                   -10.4124         12    606.328   -47.4295  (bar)
Volume                      7.92813     0.0086  0.0819152  0.0164395  (nm^3)
Density                     973.651        1.1    10.0623   -2.02709  (kg/m^3)
Enthalpy                   -8285.72        9.9    120.586    28.1691  (kJ/mol)


Volume                                   = 0.00477443 m^3/mol
Enthalpy                                 =   -8285.72 kJ/mol
Coefficient of Thermal Expansion Alpha_P = 0.000895685 (1/K)
Isothermal Compressibility Kappa         = 2.05427e-10 (J/m^3)
Adiabatic bulk modulus                   = 4.86791e+09 (m^3/J)
Heat capacity at constant pressure Cp    =    19639.5 J/mol K
Cp-Cv                                    =     5563.55 J/mol K

"""


def test_volume(get_fn):
    traj = md.load(get_fn("tip3p_300K_1ATM.xtc"), top=get_fn("tip3p_300K_1ATM.pdb"))
    v = traj.unitcell_volumes.mean()
    reference = 7.92813  # From gromacs, see above comment

    assert abs((v - reference) / reference) < 1E-3, "Volume tolerance not met!"


def test_density(get_fn):
    traj = md.load(get_fn("tip3p_300K_1ATM.xtc"), top=get_fn("tip3p_300K_1ATM.pdb"))
    rho = md.geometry.density(traj).mean()
    reference = 973.651  # From gromacs, see above comment

    assert abs((rho - reference) / reference) < 1E-3, "Density tolerance not met!"


def test_dipole_moments(get_fn):
    traj = md.load(get_fn("tip3p_300K_1ATM.xtc"), top=get_fn("tip3p_300K_1ATM.pdb"))

    charges = np.tile(tip3p_charges, traj.n_residues)

    moments0 = md.geometry.dipole_moments(traj, charges)

    # Now we screw up the molecule wholeness referencing all distances relative to atom zero.
    # E.g. [[0, 0], [0, 1], [0, 2], [0, 3]]...
    atom_indices = np.array([np.zeros(traj.n_atoms), np.arange(traj.n_atoms)], dtype='int32').T
    # Define coordinates relative to atom 0, PBC corrected.
    xyz = md.compute_displacements(traj, atom_indices, periodic=True)
    traj.xyz = xyz

    moments1 = md.geometry.dipole_moments(traj, charges)

    eq(moments0, moments1, decimal=4)


def test_static_dielectric(get_fn):
    traj = md.load(get_fn("tip3p_300K_1ATM.xtc"), top=get_fn("tip3p_300K_1ATM.pdb"))

    charges = np.tile(tip3p_charges, traj.n_residues)

    epsilon0 = md.geometry.static_dielectric(traj, charges, temperature)
    # E.g. [[0, 0], [0, 1], [0, 2], [0, 3]]...
    atom_indices = np.array([np.zeros(traj.n_atoms), np.arange(traj.n_atoms)], dtype='int32').T
    # Define coordinates relative to atom 0, PBC corrected.
    xyz = md.compute_displacements(traj, atom_indices, periodic=True)
    traj.xyz = xyz
    epsilon1 = md.geometry.static_dielectric(traj, charges, temperature)

    eq(epsilon0, epsilon1, decimal=3)

    reference = 87.1818  # From gromacs, see above comment

    assert abs((epsilon1 - reference) / reference) < 1E-3, "Dielectric tolerance not met!"


def test_kappa(get_fn):
    traj = md.load(get_fn("tip3p_300K_1ATM.xtc"), top=get_fn("tip3p_300K_1ATM.pdb"))
    kappa = md.geometry.isothermal_compressability_kappa_T(traj, temperature)
    reference = 2.05427E-10 * 1E5  # m^3 / J to 1 / bar.  Data from gromacs.  See above comment

    # 20% tolerance.
    assert abs((kappa - reference) / reference) < 2E-1, "Compressability tolerance not met!"


@pytest.mark.skip("Thermal expansion test doesn't work.")  # Not working
def test_alpha(get_fn):
    # Had some issues finding a useful unit test, so thermal_expansion_alpha_P() is currently disabled.
    # Feel free to file a pull request with a working unit test :)
    traj = md.load(get_fn("tip3p_300K_1ATM.xtc"), top=get_fn("tip3p_300K_1ATM.pdb"))

    data = pd.read_table(get_fn('tip3p_300K_1ATM.tab'), names=["timestep", "energy"], sep=r"\s*")
    energy = data.energy.values
    alpha = md.geometry.thermal_expansion_alpha_P(traj, temperature, energy)

    reference = 0.000895685  # From gromacs, see notes above.md
    eq(alpha, reference, decimal=3)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert T. McGibbon
# Contributors: Matthew Harrigan
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import tempfile, os
import mdtraj as md
from mdtraj.formats import GroTrajectoryFile
from mdtraj.testing import eq

fd, temp = tempfile.mkstemp(suffix='.gro')
os.close(fd)


def teardown_module(module):
    """remove the temporary file created by tests in this file
    this gets automatically called by pytest"""
    os.unlink(temp)


def test_read_write(get_fn):
    for t in [md.load(get_fn('4waters.pdb')),  # no unit cell
              md.load(get_fn('native2.pdb')),  # unitcell
              md.load(get_fn('frame0.pdb'))]:  # multiframe
        with GroTrajectoryFile(temp, 'w') as f:
            f.write(t.xyz, t.topology, unitcell_vectors=t.unitcell_vectors)

        with GroTrajectoryFile(temp) as f:
            xyz, time, unitcell = f.read()
            top = f.topology

        eq(xyz, t.xyz, decimal=3)
        eq(list(top.atoms), list(t.top.atoms))
        if t.unitcell_vectors is not None:
            eq(unitcell, t.unitcell_vectors)


def test_read_write_precision(get_fn):
    import random
    for t in [md.load(get_fn('4waters.pdb')),  # no unit cell
              md.load(get_fn('native2.pdb')),  # unitcell
              md.load(get_fn('frame0.pdb'))]:  # multiframe
        with GroTrajectoryFile(temp, 'w') as f:
            f.write(t.xyz, t.topology, unitcell_vectors=t.unitcell_vectors,
                    precision=random.randint(4, 10))

        with GroTrajectoryFile(temp) as f:
            xyz, time, unitcell = f.read()
            top = f.topology

        eq(xyz, t.xyz, decimal=3)
        eq(list(top.atoms), list(t.top.atoms))
        if t.unitcell_vectors is not None:
            eq(unitcell, t.unitcell_vectors)

    # Test out the automatic saver and loader with different precisions
    t = md.load(get_fn('native2.pdb'))
    t.save(temp, precision=5)
    eq(t.xyz, md.load(temp).xyz, decimal=3)


def test_no_whitespace_gro(get_fn):
    t = md.load(get_fn('v_error.gro'))
    eq(t.xyz.shape, (1, 1, 3))
    eq(t.n_atoms, 1)


def test_load(get_fn):
    for tref in [md.load(get_fn('4waters.pdb')),  # no unit cell
                 md.load(get_fn('native2.pdb'))]:  # unit cell
        with GroTrajectoryFile(temp, 'w') as f:
            f.write(tref.xyz, tref.topology, unitcell_vectors=tref.unitcell_vectors)

        t = md.load(temp)
        eq(t.xyz, tref.xyz, decimal=3)
        eq(list(t.top.atoms), list(tref.top.atoms))
        eq(t.unitcell_vectors, tref.unitcell_vectors)

    #if there are 2 residues with same renum but different resname
    t = md.load(get_fn('two_residues_same_resnum.gro'))
    eq(t.n_residues, 2)

def test_load_with_given_top(get_fn):
    gro = get_fn('frame0.gro')
    t = md.load(gro, top=gro)
    tref = md.load(gro)
                 
    eq(t.xyz, tref.xyz)
    eq(t.top, tref.top)
    eq(list(t.top.atoms), list(tref.top.atoms))
    eq(t.unitcell_vectors, tref.unitcell_vectors)




def test_against_gmx(get_fn, tmpdir):
    t1 = md.load(get_fn('frame0.pdb'))

    # generated by converting frame0.pdb to gro with g_trajconv
    t2 = md.load(get_fn('frame0.gro'))

    fn = "{}/temp.gro".format(tmpdir)
    t1.save(fn)
    t3 = md.load(fn)

    eq(t1.xyz, t2.xyz)
    eq(t1.xyz, t3.xyz)
    eq(t1.time, t2.time)
    eq(t1.time, t3.time)
    eq(t1.unitcell_vectors, t2.unitcell_vectors)
    eq(t1.unitcell_vectors, t3.unitcell_vectors)



from mdtraj import load
from mdtraj.testing import eq

def test_load_single(get_fn):
    # Just check for any raised errors coming from loading a single file.
    load(get_fn('frame0.pdb'))


def test_load_single_list(get_fn):
    # See if a single-element list of files is successfully loaded.
    load([get_fn('frame0.pdb')])


def test_load_many_list(get_fn):
    # See if a multi-element list of files is successfully loaded.
    single = load(get_fn('frame0.pdb'))
    double = load(2 * [get_fn('frame0.pdb')], discard_overlapping_frames=False)
    assert 2 * single.n_frames == double.n_frames

def test_load_atom_indices_multiple_files(get_fn):
    ref_t = load(get_fn('native.pdb'))
    t = load([get_fn('native.pdb')]*2, atom_indices=[0])

    eq(t.topology, ref_t.topology.subset([0]))



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import tempfile, os
import numpy as np
import mdtraj as md
from mdtraj.formats import MDCRDTrajectoryFile
from mdtraj.testing import eq

fd, temp = tempfile.mkstemp(suffix='.mdcrd')


def teardown_module(module):
    """remove the temporary file created by tests in this file
    this gets automatically called by pytest"""
    os.close(fd)
    os.unlink(temp)


def test_read_0(get_fn):
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrd'), n_atoms=22) as f:
        xyz, box = f.read()
        assert box is None
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrdbox'), n_atoms=22) as f:
        xyz2, box = f.read()
        eq(box.shape, (1002, 3))
        eq(xyz, xyz2)


def test_read_1(get_fn):
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrd'), n_atoms=22) as f:
        xyz, _ = f.read()
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrd'), n_atoms=22) as f:
        xyz3, _ = f.read(stride=3)

    eq(xyz[::3], xyz3)


def test_read_write_0():
    xyz = 10 * np.random.randn(100, 11, 3)
    with MDCRDTrajectoryFile(temp, mode='w') as f:
        f.write(xyz)
    with MDCRDTrajectoryFile(temp, n_atoms=11) as f:
        xyz2, _ = f.read()

    eq(_, None)
    eq(xyz, xyz2, decimal=3)


def test_read_write_1():
    xyz = 10 * np.random.randn(100, 11, 3)
    box = np.random.randn(100, 3)
    with MDCRDTrajectoryFile(temp, mode='w') as f:
        f.write(xyz, box)
    with MDCRDTrajectoryFile(temp, n_atoms=11) as f:
        xyz2, box2 = f.read()

    eq(box, box2, decimal=3)
    eq(xyz, xyz2, decimal=3)


def test_read_write_2(get_fn):
    pdb = md.load(get_fn('1bpi.pdb'))
    pdb.save(temp)

    t = md.load(temp, top=pdb.topology)

    eq(t.xyz, pdb.xyz)
    eq(t.unitcell_vectors, pdb.unitcell_vectors)


def test_multiread(get_fn):
    reference = md.load(get_fn('frame0.mdcrd'), top=get_fn('native.pdb'))
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrd'), n_atoms=22) as f:
        xyz0, box0 = f.read(n_frames=1)
        xyz1, box1 = f.read(n_frames=1)

    eq(reference.xyz[0], xyz0[0] / 10)
    eq(reference.xyz[1], xyz1[0] / 10)


def test_seek(get_fn):
    reference = md.load(get_fn('frame0.mdcrd'), top=get_fn('native.pdb'))
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrd'), n_atoms=22) as f:
        f.seek(1)
        eq(1, f.tell())
        xyz1, box1 = f.read(n_frames=1)
        eq(reference.xyz[1], xyz1[0] / 10)

        f.seek(10)
        eq(10, f.tell())
        xyz10, box10 = f.read(n_frames=1)
        eq(reference.xyz[10], xyz10[0] / 10)
        eq(11, f.tell())

        f.seek(-8, 1)
        xyz3, box3 = f.read(n_frames=1)
        eq(reference.xyz[3], xyz3[0] / 10)

        f.seek(4, 1)
        xyz8, box8 = f.read(n_frames=1)
        eq(reference.xyz[8], xyz8[0] / 10)


def test_atom_indices_0(get_fn):
    atom_indices = np.arange(10)
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrd'), n_atoms=22) as f:
        xyz0, box0 = f.read(n_frames=1)
    with MDCRDTrajectoryFile(get_fn('frame0.mdcrd'), n_atoms=22) as f:
        xyz1, box1 = f.read(n_frames=1, atom_indices=atom_indices)

    eq(xyz0[:, atom_indices], xyz1)


def test_atom_indices_1(get_fn):
    atom_indices = np.arange(10)
    top = md.load(get_fn('native.pdb'))
    t0 = md.load(get_fn('frame0.mdcrd'), top=top)
    t1 = md.load(get_fn('frame0.mdcrd'), top=top, atom_indices=atom_indices)

    eq(t0.xyz[:, atom_indices], t1.xyz)



from __future__ import print_function, division, absolute_import

import os
import io
import bz2
import gzip
from mdtraj.utils.six import PY2, StringIO


def open_maybe_zipped(filename, mode, force_overwrite=True):
    """Open a file in text (not binary) mode, transparently handling
    .gz or .bz2 compresssion, with utf-8 encoding.

    Parameters
    ----------
    filename : str
        Path to file. Compression will be automatically detected if
        the filename ends in .gz or .bz2.
    mode : {'r', 'w'}
        Mode in which to open file
    force_overwrite : bool, default=True
        If 'w', should we overwrite the file if something with `filename`
        already exists?

    Returns
    -------
    handle : file
        Open file handle.
    """
    _, extension = os.path.splitext(str(filename).lower())
    if mode == 'r':
        if extension == '.gz':
            with gzip.GzipFile(filename, 'r') as gz_f:
                return StringIO(gz_f.read().decode('utf-8'))
        elif extension == '.bz2':
            with bz2.BZ2File(filename, 'r') as bz2_f:
                return StringIO(bz2_f.read().decode('utf-8'))
        else:
            return open(filename, 'r')
    elif mode == 'w':
        if os.path.exists(filename) and not force_overwrite:
            raise IOError('"%s" already exists' % filename)
        if extension == '.gz':
            if PY2:
                return gzip.GzipFile(filename, 'w')
            binary_fh = gzip.GzipFile(filename, 'wb')
            return io.TextIOWrapper(binary_fh, encoding='utf-8')
        elif extension == '.bz2':
            if PY2:
                return bz2.BZ2File(filename, 'w')
            binary_fh = bz2.BZ2File(filename, 'wb')
            return io.TextIOWrapper(binary_fh, encoding='utf-8')
        else:
            return open(filename, 'w')
    else:
        raise ValueError('Invalid mode "%s"' % mode)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: TJ Lane
# Contributors: Robert McGibbon, Jason Swails
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import mdtraj as md
from mdtraj.testing import eq
from mdtraj.formats import prmtop


def test_load_prmtop(get_fn):
    top = prmtop.load_prmtop(get_fn('alanine-dipeptide-implicit.prmtop'))
    ref_top = md.load(get_fn('native.pdb')).topology
    eq(top, ref_top)


def test_load_binpos_w_prmtop(get_fn):
    traj = md.load(get_fn('alanine.binpos'), top=get_fn('alanine-dipeptide-implicit.prmtop'))
    ref_traj = md.load(get_fn('native.pdb'))

    eq(traj.topology, ref_traj.topology)
    eq(traj.xyz, ref_traj.xyz)

def test_load_chamber_prmtop(get_fn):
    top = prmtop.load_prmtop(get_fn('ala3_chamber.parm7'))
    eq(top.n_atoms, 33)
    eq(top.n_residues, 3)
    eq(top.n_bonds, 32)
    eq(top.n_chains, 1)



cdef extern from "include/binposplugin.h":
    
    void *open_binpos_read(char *path, char *filetype, int *natoms)
    int read_next_timestep(void *v, int natoms, molfile_timestep_t *ts)
    void close_file_read(void *v)

    void *open_binpos_write(char *path, char *filetype, int natoms)
    int write_timestep(void *v, molfile_timestep_t *ts)
    void close_file_write(void *v)
    int seek_timestep(void* v, long int offset, int origin)
    long int tell_timestep(void* v)

cdef extern from "include/molfile_plugin.h":
    ctypedef struct molfile_timestep_t:
      float *coords  # coordinates of all atoms, arranged xyzxyzxyz
      float *velocities  # space for velocities of all atoms; same layout
      float A, B, C, alpha, beta, gamma




##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors: Lee-Ping Wang
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


##############################################################################
# Imports
##############################################################################
import cython
import numpy as np
from mdtraj.utils import ensure_type
import scipy.spatial.distance

cimport numpy as np
from libcpp.vector cimport vector
from libc.stdio cimport printf
from cpython cimport bool
from cython.parallel cimport prange
np.import_array()
assert sizeof(np.int32_t) == sizeof(int)


##############################################################################
# External c/cpp function declarations
##############################################################################
cdef extern from "include/fancy_index.hpp":
    cdef extern void fancy_index2d(const float* A, int nx, int ny,
        const int* indx, int nindx, const int* indy, int nindy, float* out) nogil
cdef extern float msd_atom_major(int nrealatoms, int npaddedatoms,  float* a,
    float* b, float G_a, float G_b, int computeRot, float rot[9]) nogil
cdef extern float rot_atom_major(const int n_atoms, float* a, const float rot[9]) nogil
cdef extern void sgemm33(const float A[9], const float B[9], float out[9]) nogil
cdef extern void inplace_center_and_trace_atom_major(float* coords, float* traces,
    const int n_frames, const int n_atoms) nogil
cdef extern from "include/Munkres.h":
    cdef cppclass Munkres:
        Munkres()
        void solve(double* icost, int* answer, int m, int n)
cdef extern from "include/euclidean_permutation.hpp":
   vector[int] euclidean_permutation(float* target, float* reference,
       int n_atoms, int n_dims, vector[vector[int]]& permute_groups) nogil
cdef extern from "math.h":
    float sqrtf(float x) nogil


##############################################################################
# Functions
##############################################################################

def lprmsd(target, reference, int frame=0, atom_indices=None, permute_groups=None,
           bool parallel=True, bool superpose=False):
    """lprmsd(target, reference, frame=0, atom_indices=None, permute_groups=None, bool parallel=True, bool superpose=False)

    Compute Linear-Programming Root-Mean-Squared Deviation (LP-RMSD) of all
    conformations in target to a reference conformation. The LP-RMSD is the
    minimum root-mean squared deviation between two sets of points, minimizing
    over both the rotational/translational degrees of freedom AND the label
    correspondences between points in the target and reference conformations.
    This means that it can be used meaningfully with atoms with exchange
    symmetry like, like multiple water molecules.

    Notes
    -----
    The optimization procedure used does necessary not find the rotation/mapping
    for the global minimum of the means squared deviation. It does usually find
    a *pretty good* solution though, using a 3 step procedure. For each frame in
    the target, (1) align the target reference frame using only the
    distinguishable atoms (2) using this fixed rotation matrix, find the optimal
    mapping for the labels in the permute groups (3) using this mapping,
    recompute a new optimal rotation matrix to align the frame, and get the
    final RMSD.

    Parameters
    ----------
    target : md.Trajectory
        For each conformation in this trajectory, compute the RMSD to
        a particular 'reference' conformation in another trajectory
        object.
    reference : md.Trajectory
        The object containing the reference conformation to measure distances
        to.
    frame : int
        The index of the conformation in `reference` to measure
        distances to.
    atom_indices : array_like, or None
        The indices of the atoms to use in the RMSD calculation. If not
        supplied, all atoms will be used.
    permute_groups : list of array_like, or None
        A list of groups of permutable atoms. Each element in permute_groups
        is an array of indices containing atoms whose labels can be mutually
        exchanged within the group. All indices in permute_groups must be a
        subset of atom_indices. The default, None, corresponds to a single group
        containing all the atoms in atom_indices.
    parallel : bool
        Use OpenMP to calculate each of the RMSDs in parallel over multiple
        cores.
    superpose : bool
        Modify the trajectory ``target`` in-place to superpose each frame on
        reference.

    Returns
    -------
    lprmsds : np.ndarray, shape=(target.n_frames,)
        A 1-D numpy array of the linear programming root-mean-square deviations
        from each of the conformations in `target` to the `frame`-th
        conformation in `reference`.
    """
    _validate_shapes(target, reference, frame)
    atom_indices = _validate_atom_indices(atom_indices, target.xyz.shape[1])
    permute_groups = _validate_permute_groups(permute_groups, atom_indices)

    # these are the indices of the permute atoms inside the coords array after
    # selecting only the atom indices.
    # i.e. [xyz[atom_indices][i] for in permute_groups_rel]
    cdef vector[vector[int]] permute_groups_stl
    for pgroup in permute_groups:
         permute_groups_stl.push_back(np.searchsorted(atom_indices, pgroup))
    dis_indices = np.setdiff1d(
            np.arange(len(atom_indices)), np.concatenate(permute_groups_stl))
    # print('dis indices', dis_indices)

    cdef int n_atoms_total = target.xyz.shape[1]
    cdef int superpose_ = superpose
    cdef np.ndarray[ndim=3, dtype=np.float32_t, mode='c'] target_xyz = np.asarray(target.xyz, order='c')
    cdef np.ndarray[ndim=1, dtype=int, mode='c'] atom_indices_ = np.asarray(atom_indices, dtype=np.int32)
    cdef np.ndarray[ndim=1, dtype=int, mode='c'] dis_indices_ =  np.asarray(dis_indices, dtype=np.int32)

    # get the all the coords in atom_indices in target and ref.
    # center them and compute the g values
    cdef int i
    cdef double msd
    cdef float ref_g, target_g
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] target_xyz_frame
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] target_xyz_frame2
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] ref_xyz_frame
    cdef int target_n_frames = target.xyz.shape[0]
    cdef int n_atoms = len(atom_indices)
    ref_xyz_frame = np.array(reference.xyz[frame, atom_indices, :], dtype=np.float32, copy=True)
    target_xyz_frame = np.finfo(np.float32).max * np.ones_like(ref_xyz_frame)
    target_xyz_frame2 = np.empty_like(target_xyz_frame)
    inplace_center_and_trace_atom_major(&ref_xyz_frame[0, 0], &ref_g, 1, n_atoms)


    # get only the distinguishable atoms in target and ref, center them
    # and compute the g values
    cdef float ref_g_dis, target_g_dis
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] target_xyz_frame_dis
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] ref_xyz_frame_dis
    cdef int n_atoms_dis = len(dis_indices)
    target_xyz_frame_dis = np.empty((n_atoms_dis, 3), dtype=np.float32)
    if n_atoms_dis > 0:
        # note: the indexing here is subtpe -- we're using atom_indices[dis_indices] since
        # dis_indices have different semantics from atom_indices
        ref_xyz_frame_dis = np.array(reference.xyz[frame, atom_indices[dis_indices], :], copy=True, order='c')
        inplace_center_and_trace_atom_major(&ref_xyz_frame_dis[0, 0], &ref_g_dis, 1, n_atoms_dis)


    cdef vector[int] mapping
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] rot1 = np.eye(3, dtype=np.float32)
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] rot2 = np.eye(3, dtype=np.float32)
    cdef np.ndarray[ndim=2, dtype=np.float32_t, mode='c'] rot3 = np.eye(3, dtype=np.float32)
    cdef np.ndarray[ndim=1, dtype=np.float32_t, mode='c'] distances = np.zeros(target_n_frames, dtype=np.float32)

    #for i in prange(target_n_frames, nogil=True):
    for i in range(target_n_frames):
        fancy_index2d(&target_xyz[i,0,0], n_atoms_total, 3, <int*> &atom_indices_[0],
                      n_atoms, NULL, 0, &target_xyz_frame[0, 0])
        inplace_center_and_trace_atom_major(&target_xyz_frame[0, 0], &target_g, 1, n_atoms)

        # compute rotation matrix on distinguishable atoms
        if (n_atoms_dis > 0):
            # subsample the target_xyz_frame to get just the distinguishable atoms
            # target_xyz_frame has already been "sampled down" by atom_indices, so we
            # can now apply dis_indices
            fancy_index2d(&target_xyz_frame[0,0], n_atoms, 3, <int*> &dis_indices_[0],
                          n_atoms_dis, NULL, 0, &target_xyz_frame_dis[0, 0])
            inplace_center_and_trace_atom_major(&target_xyz_frame_dis[0,0], &target_g_dis, 1, n_atoms_dis)

            # get `rot1`, the rotation matrix thats optimal for the distinguishable indices
            msd_atom_major(n_atoms_dis, n_atoms_dis,
                &target_xyz_frame_dis[0, 0], &ref_xyz_frame_dis[0, 0],
                target_g_dis, ref_g_dis, 1, &rot1[0, 0])

            # apply rot1 to all the atom_indices
            rot_atom_major(n_atoms, &target_xyz_frame[0, 0], &rot1[0, 0])

        # compute the optimal remapping of the indices
        mapping = euclidean_permutation(&ref_xyz_frame[0, 0], &target_xyz_frame[0, 0], n_atoms, 3, permute_groups_stl)
        # and remap the indices -- i.e.   target_xyz_frame2 = target_xyz_frame[mapping]
        fancy_index2d(&target_xyz_frame[0, 0], n_atoms, 3, &mapping[0], n_atoms, NULL, 0, &target_xyz_frame2[0, 0])

        # msd = ((target_xyz_frame2 - ref_xyz_frame)**2).sum(1).mean(0)
        # then using these remapped indices, compute the rmsd, with a new rotation
        if superpose_:
            msd = msd_atom_major(n_atoms, n_atoms, &target_xyz_frame2[0, 0], &ref_xyz_frame[0, 0], target_g, ref_g, 1, &rot2[0, 0])
            inplace_center_and_trace_atom_major(&target_xyz[i, 0, 0], NULL, 1, n_atoms_total)
            sgemm33(&rot1[0,0], &rot2[0,0], &rot3[0,0])
            rot_atom_major(n_atoms_total, &target_xyz[i, 0, 0], &rot3[0, 0])
        else:
            msd = msd_atom_major(n_atoms, n_atoms, &ref_xyz_frame[0, 0], &target_xyz_frame2[0, 0], target_g, ref_g, 0, NULL)



        distances[i] = sqrtf(msd)

    if superpose_:
        target.xyz = target_xyz

    return distances


def _validate_atom_indices(atom_indices, n_atoms):
    if atom_indices is None:
        atom_indices = np.arange(n_atoms, dtype=int)
    else:
        atom_indices = ensure_type(np.unique(atom_indices), dtype=int,
                                   ndim=1, name='atom_indices', warn_on_cast=False)
        if not np.all((atom_indices >= 0) * (atom_indices < n_atoms) * (atom_indices < n_atoms)):
            raise ValueError("atom_indices must be valid positive indices")
    return atom_indices


def _validate_shapes(target, reference, frame):
    assert (target.xyz.ndim == 3) and (reference.xyz.ndim == 3) and \
           (target.xyz.shape[2]) == 3 and (reference.xyz.shape[2] == 3)
    if not (target.xyz.shape[1]  == reference.xyz.shape[1]):
        raise ValueError("Input trajectories must have same number of atoms. "
                         "found %d and %d." % (target.xyz.shape[1], reference.xyz.shape[1]))
    if frame >= reference.xyz.shape[0]:
        raise ValueError("Cannot calculate RMSD of frame %d: reference has "
                         "only %d frames." % (frame, reference.xyz.shape[0]))


def _validate_permute_groups(permute_groups, atom_indices):
    if permute_groups is None:
        return [atom_indices]

    permute_groups = [ensure_type(np.unique(group), dtype=int, ndim=1,
                         warn_on_cast=False, name='permute_groups[%d]' % i) for \
                         i, group in enumerate(permute_groups)]
    for pgroup in permute_groups:
        if len(np.setdiff1d(pgroup, atom_indices)) > 0:
            raise ValueError('The elements in each permute group must be a subset of atom_indices')

    all_permutable_atoms = np.concatenate(permute_groups)
    if not np.all(np.unique(all_permutable_atoms) == np.sort(all_permutable_atoms)):
        raise ValueError('permute_groups must be mutually disjoint sets')

    return permute_groups


@cython.boundscheck(False)
def _munkres(np.ndarray[np.double_t, ndim=2, mode="c"] A not None):
    """_munkres(A)

    Calculate the minimum cost assignment of a cost matrix, A

    Parameters
    ----------
    A : np.ndarray, dtype=np.double, ndim=2

    Returns
    -------
    assignments : np.ndarray, ndim=2, dtype=int32
        Boolean array with shape equal to the shape of A. assignments[i,j] == 1
        for an assignment, and 0 for a non-assignment

    Examples
    -------
    >>> _munkres(np.array([[7, 4, 3], [6, 8, 5], [9, 4, 4]], dtype=np.double))
    [[0 0 1],
     [1 0 0],
     [0 1 0]]
    """
    cdef int x = A.shape[0]
    cdef int y = A.shape[1]
    cdef np.ndarray[ndim=2, dtype=np.int32_t, mode='c'] rslt

    rslt = np.zeros(shape=(x,y), dtype=np.int32, order='c')
    cdef Munkres* munk = new Munkres()
    munk.solve(<double *> A.data, <int *> rslt.data, x, y)
    del munk

    return rslt



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import mdtraj as md
import numpy as np
from mdtraj.geometry.alignment import rmsd_qcp, compute_average_structure
from mdtraj.testing import eq

np.random.seed(52)


def test_trajectory_rmsd(get_fn):
    t = md.load(get_fn('traj.h5'))
    for parallel in [True, False]:
        calculated = md.rmsd(t, t, 0, parallel=parallel)
        reference = np.zeros(t.n_frames)
        for i in range(t.n_frames):
            reference[i] = rmsd_qcp(t.xyz[0], t.xyz[i])

        eq(calculated, reference, decimal=3)


def test_precentered_1(get_fn):
    # test rmsd against the numpy version, using the same trajectory
    # as target and reference
    t1 = md.load(get_fn('traj.h5'), stride=10)
    t2 = md.load(get_fn('traj.h5'), stride=10)
    # don't center t1, and use it without precentered
    # explicitly center t2, and use *with* precentered

    for parallel in [True, False]:
        t2.center_coordinates()
        eq(t1.n_frames, t2.n_frames)
        for i in range(t1.n_frames):
            ref = np.zeros(t1.n_frames)
            for j in range(t1.n_frames):
                ref[j] = rmsd_qcp(t1.xyz[j], t1.xyz[i])
            val1 = md.rmsd(t1, t1, i, parallel=parallel, precentered=False)
            val2 = md.rmsd(t2, t2, i, parallel=parallel, precentered=True)

            eq(ref, val1, decimal=3)
            eq(val1, val2)


def test_precentered_2(get_fn):
    # test rmsd against the numpy version, using the difference
    # trajectories as target and reference
    t1_a = md.load(get_fn('traj.h5'), stride=10)
    t2_a = md.load(get_fn('traj.h5'), stride=10)
    t1_b = md.load(get_fn('traj.h5'), stride=10)
    t2_b = md.load(get_fn('traj.h5'), stride=10)
    # don't center t1, and use it without precentered
    # explicitly center t2, and use *with* precentered

    t2_a.center_coordinates()
    t2_b.center_coordinates()

    for parallel in [True, False]:
        for i in range(t1_b.n_frames):
            ref = np.zeros(t1_a.n_frames)
            for j in range(t1_a.n_frames):
                ref[j] = rmsd_qcp(t1_a.xyz[j], t1_b.xyz[i])
            val1 = md.rmsd(t1_a, t1_b, i, parallel=parallel, precentered=False)
            val2 = md.rmsd(t2_a, t2_b, i, parallel=parallel, precentered=True)

            eq(ref, val1, decimal=3)
            eq(val1, val2, decimal=4)


def test_superpose_0(get_fn):
    t1 = md.load(get_fn('traj.h5'))
    reference_rmsd = md.rmsd(t1, t1, 0)

    t1.superpose(t1, 0)
    displ_rmsd = np.zeros(t1.n_frames)
    for i in range(t1.n_frames):
        delta = t1.xyz[i] - t1.xyz[0]
        displ_rmsd[i] = (delta ** 2.0).sum(1).mean() ** 0.5

    eq(reference_rmsd, displ_rmsd, decimal=5)


def test_superpose_1():
    # make one frame far from the origin
    reference = md.Trajectory(xyz=np.random.randn(1, 100, 3) + 100, topology=None)
    reference_xyz = reference.xyz.copy()

    for indices in [None, np.arange(90)]:
        # make another trajectory in a similar rotational state
        query = md.Trajectory(xyz=reference.xyz + 0.01 * np.random.randn(*reference.xyz.shape), topology=None)
        query.superpose(reference, 0, atom_indices=indices)
        assert eq(reference.xyz, reference_xyz)

        new_centers = np.mean(query.xyz[0], axis=1)
        assert 80 < new_centers[0] < 120
        assert 80 < new_centers[1] < 120
        assert 80 < new_centers[2] < 120


def test_superpose_2():
    t1 = md.Trajectory(xyz=np.random.randn(1, 100, 3) + 100, topology=None)
    t2 = md.Trajectory(xyz=np.random.randn(1, 100, 3) + 100, topology=None)
    t2_copy = t2.xyz.copy()

    t1.superpose(t2)
    t1.superpose(t2, atom_indices=[1, 2, 3, 4, 5, 6, 7])

    # make sure that superposing doesn't alter the reference traj
    eq(t2.xyz, t2_copy)


def test_superpose_refinds():
    # make one frame far from the origin
    normal = np.random.randn(1, 100, 3)
    normal_xyz = normal.copy()

    flipped = np.zeros_like(normal)
    flipped[:, :50, :] = normal[:, 50:, :]
    flipped[:, 50:, :] = normal[:, :50, :]
    flipped_xyz = flipped.copy()

    normal = md.Trajectory(xyz=normal, topology=None)
    flipped = md.Trajectory(xyz=flipped, topology=None)

    normal.superpose(flipped, atom_indices=np.arange(0, 50), ref_atom_indices=np.arange(50, 100))
    eq(normal.xyz, normal_xyz)

    flipped.superpose(normal, atom_indices=np.arange(50, 100), ref_atom_indices=np.arange(0, 50))
    eq(flipped.xyz, flipped_xyz)

    normal.superpose(flipped)
    assert not np.allclose(normal.xyz, normal_xyz)


def test_rmsd_atom_indices(get_fn):
    native = md.load(get_fn('native.pdb'))
    t1 = md.load(get_fn('traj.h5'))

    atom_indices = np.arange(10)
    dist1 = md.rmsd(t1, native, atom_indices=atom_indices)

    t2 = md.load(get_fn('traj.h5'))
    t2.restrict_atoms(atom_indices)
    native.restrict_atoms(atom_indices)
    dist2 = md.rmsd(t2, native)

    eq(dist1, dist2)


def test_rmsd_ref_ainds(get_fn):
    native = md.load(get_fn('native.pdb'))
    t1 = md.load(get_fn('traj.h5'))

    atom_indices = np.arange(10)
    dist1 = md.rmsd(t1, native, atom_indices=atom_indices,
                    ref_atom_indices=atom_indices)

    bad_atom_indices = np.arange(10, 20)
    t2 = md.load(get_fn('traj.h5'))
    dist2 = md.rmsd(t2, native, atom_indices=atom_indices,
                    ref_atom_indices=bad_atom_indices)

    assert np.all(dist2 > dist1)


def test_average_structure(get_fn):
    traj = md.load(get_fn('frame0.dcd'), top=get_fn('frame0.pdb'))
    average = compute_average_structure(traj.xyz)

    # The mean RMSD to the average structure should be less than to any individual frame.
    sum1 = 0
    sum2 = 0
    for i in range(traj.n_frames):
        sum1 += rmsd_qcp(traj.xyz[0], traj.xyz[i])
        sum2 += rmsd_qcp(average, traj.xyz[i])
    assert sum2 < sum1

def test_trajectory_rmsf(get_fn):
    t = md.load(get_fn('traj.h5'))
    for parallel in [True, False]:
        calculated = md.rmsf(t, t, 0, parallel=parallel)
        t.superpose(t, 0)
        avg_xyz = np.average(t.xyz, axis=0)
        reference = np.sqrt(3*np.mean((t.xyz - avg_xyz)**2, axis=(0, 2)))
        assert np.sum(np.abs(calculated)) > 0 # check trivial error
        eq(calculated, reference, decimal=3)

def test_trajectory_rmsf_aligned(get_fn):
    t = md.load(get_fn('traj.h5'))
    for parallel in [True, False]:
        # testing different set of atoms for alignment and RMSF calculation
        atom_indices = range(int(t.n_atoms/2))
        rmsf_indices = range(int(t.n_atoms/2), t.n_atoms)
        t.superpose(t, 99, atom_indices=atom_indices, parallel=False)
        calculated = md.rmsf(t, None, atom_indices=rmsf_indices, parallel=parallel)
        avg_xyz = np.average(t.xyz, axis=0)
        reference = np.sqrt(3*np.mean((t.xyz - avg_xyz)**2, axis=(0, 2)))[rmsf_indices]
        assert np.sum(np.abs(calculated)) > 0 # check trivial error
        eq(calculated, reference, decimal=3)

def test_rmsd_atom_indices_vs_ref_indices():
    n_frames = 1
    n_atoms_1 = 1
    n_atoms_2 = 2

    top_1 = md.Topology()
    top_1.add_chain()
    top_1.add_residue('RS2', top_1.chain(0))
    top_1.add_atom('A2', 'H', top_1.residue(0))

    top_2 = md.Topology()
    top_2.add_chain()
    top_2.add_residue('RS1', top_2.chain(0))
    top_2.add_atom('A1', 'H', top_2.residue(0))
    top_2.add_chain()
    top_2.add_residue('RS2', top_2.chain(1))
    top_2.add_atom('A2', 'H', top_2.residue(1))
    # here the 2nd chain in the top_2 is rmsd-compatible to the one in the top_1 so we should be able to compute rsmd between them.

    trj_1 = md.Trajectory(np.random.RandomState(0).randn(n_frames, n_atoms_1, 3), top_1)
    trj_2 = md.Trajectory(np.random.RandomState(0).randn(n_frames, n_atoms_2, 3), top_2)

    md.rmsd(trj_1, trj_2, atom_indices=[0], ref_atom_indices=[1])
    md.rmsd(trj_2, trj_1, atom_indices=[1], ref_atom_indices=[0])
    # is this don't fail then it's good no matter the result



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import mdtraj as md
import numpy as np
from mdtraj import element
from mdtraj.geometry.sasa import _ATOMIC_RADII
from mdtraj.testing import eq

# set up a mock topology with 1 atom
topology1 = md.Topology()
topology1.add_atom('H', element.hydrogen, topology1.add_residue('res', topology1.add_chain()))

# set up a mock topology with two atoms
topology2 = md.Topology()
_res2 = topology2.add_residue('res', topology2.add_chain())
topology2.add_atom('H', element.hydrogen, _res2)
topology2.add_atom('H', element.hydrogen, _res2)

# set up a mock topologies with a single chlorine and sulfur atom
topology3 = md.Topology()
topology4 = md.Topology()
topology3.add_atom('Cl', element.chlorine, topology3.add_residue('res', topology3.add_chain()))
topology4.add_atom('S', element.sulfur, topology4.add_residue('res', topology4.add_chain()))


def test_sasa_0():
    # make one atom at the origin
    traj = md.Trajectory(xyz=np.zeros((1, 1, 3)), topology=topology1)

    probe_radius = 0.14
    calc_area = np.sum(md.geometry.shrake_rupley(traj, probe_radius=probe_radius))
    true_area = 4 * np.pi * (_ATOMIC_RADII['H'] + probe_radius) ** 2

    np.testing.assert_approx_equal(calc_area, true_area)


def test_sasa_1():
    # two atoms
    traj = md.Trajectory(xyz=np.zeros((1, 2, 3)), topology=topology2)

    probe_radius = 0.14
    true = 4 * np.pi * (_ATOMIC_RADII['H'] + probe_radius) ** 2

    # when atoms are closer than 2e-5, there seems to be a bug.
    # note that you should never actually have a case where atoms are this close
    # but nonetheless I'm adding a check for this in the implementation -- to make
    # it crash if the atoms are too close, as opposed to giving you wrong results
    separations = np.linspace(2.0e-5, probe_radius * 2 + _ATOMIC_RADII['H'] * 2, 10)
    areas = np.zeros_like(separations)

    # check the sasa as we vary the separation
    for i, sep in enumerate(separations):
        traj.xyz[0, 0, 1] = sep
        areas[i] = np.sum(md.geometry.shrake_rupley(traj, probe_radius=probe_radius))

    np.testing.assert_approx_equal(areas[0], true, significant=3)
    np.testing.assert_approx_equal(areas[-1], 2 * true)
    # make sure that areas is increasing
    np.testing.assert_array_less(areas[0:8], areas[1:9])


def test_sasa_2(get_fn):
    t = md.load(get_fn('frame0.h5'))
    val1 = np.sum(md.geometry.shrake_rupley(t[0]))  # calculate only frame 0
    val2 = np.sum(md.geometry.shrake_rupley(t)[0])  # calculate on all frames
    true_frame_0_sasa = 3.5556480884552

    np.testing.assert_approx_equal(true_frame_0_sasa, val1)
    np.testing.assert_approx_equal(true_frame_0_sasa, val2)


def test_sasa_3(get_fn):
    traj_ref = np.loadtxt(get_fn('gmx_sasa.dat'))
    traj = md.load(get_fn('frame0.h5'))
    traj_sasa = md.geometry.shrake_rupley(traj, probe_radius=0.14, n_sphere_points=960)

    # the algorithm used by gromacs' g_sas is slightly different than the one
    # used here, so the results are not exactly the same
    np.testing.assert_array_almost_equal(traj_sasa, traj_ref, decimal=1)


def test_sasa_4(get_fn):
    def _test_atom_group(t, value):
        sasa = md.shrake_rupley(t, mode='atom')
        rids = np.array([a.residue.index for a in t.top.atoms])

        for i, rid in enumerate(np.unique(rids)):
            mask = (rids == rid)
            eq(value[:, i], np.sum(sasa[:, mask], axis=1))

    t = md.load(get_fn('frame0.h5'))
    value = md.shrake_rupley(t, mode='residue')
    assert value.shape == (t.n_frames, t.n_residues)
    _test_atom_group(t, value)

    # scramle the order of the atoms, and which residue each is a
    # member of
    df, bonds = t.top.to_dataframe()
    df['resSeq'] = np.random.permutation(df['resSeq'])
    df['resName'] = df['resSeq']
    t.top = md.Topology.from_dataframe(df, bonds)

    value = md.shrake_rupley(t, mode='residue')
    _test_atom_group(t, value)


def test_sasa_5():
    # Test if you can change atom radii effectively without breaking things
    traj_chlorine = md.Trajectory(xyz=np.zeros((1, 1, 3)), topology=topology3)
    traj_sulfur = md.Trajectory(xyz=np.zeros((1, 1, 3)), topology=topology4)

    probe_radius = 0.14
    change_radii = {'Cl': 0.175}
    calc_area_chlorine = np.sum(md.geometry.shrake_rupley(traj_chlorine,
                                                          probe_radius=probe_radius,
                                                          change_radii=change_radii))
    calc_area_sulfur = np.sum(md.geometry.shrake_rupley(traj_sulfur,
                                                        probe_radius=probe_radius,
                                                        change_radii=change_radii))
    true_area_chlorine = 4 * np.pi * (0.175 + probe_radius) ** 2
    # This atom should remain unchanged
    true_area_sulfur = 4 * np.pi * (_ATOMIC_RADII['S'] + probe_radius) ** 2
    # And calc_area_chlorine should be different from this value
    prev_area_chlorine = 4 * np.pi * (_ATOMIC_RADII['Cl'] + probe_radius) ** 2

    np.testing.assert_approx_equal(calc_area_chlorine, true_area_chlorine)
    # Also ensure that rest of dict is not gettting deleted
    np.testing.assert_approx_equal(calc_area_sulfur, true_area_sulfur)
    # Finally, ensure that we are not changing _ATOMIC_RADII
    assert prev_area_chlorine != calc_area_chlorine



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

from mdtraj.formats import NetCDFTrajectoryFile
import os, tempfile
import numpy as np
import mdtraj as md
import subprocess
from mdtraj.testing import eq
import pytest

from distutils.spawn import find_executable

needs_cpptraj = pytest.mark.skipif(
    find_executable('cpptraj') is None,
    reason="This test requires cpptraj from AmberTools to be installed "
           "(http://ambermd.org)"
)

fd, temp = tempfile.mkstemp(suffix='.nc')
fd2, temp2 = tempfile.mkstemp(suffix='.nc')


def teardown_module(module):
    """remove the temporary file created by tests in this file
    this gets automatically called by pytest"""
    os.close(fd)
    os.close(fd2)
    os.unlink(temp)
    os.unlink(temp2)


def test_read_after_close(get_fn):
    f = NetCDFTrajectoryFile(get_fn('mdcrd.nc'))
    assert eq(f.n_atoms, 223)
    assert eq(f.n_frames, 101)

    f.close()

    # should be an IOError if you read a file that's closed
    with pytest.raises(IOError):
        f.read()


def test_shape(get_fn):
    xyz, time, boxlength, boxangles = NetCDFTrajectoryFile(get_fn('mdcrd.nc')).read()

    assert eq(xyz.shape, (101, 223, 3))
    assert eq(time.shape, (101,))
    assert eq(boxlength, None)
    assert eq(boxangles, None)


def test_read_chunk_1(get_fn):
    with NetCDFTrajectoryFile(get_fn('mdcrd.nc')) as f:
        a, b, c, d = f.read(10)
        e, f, g, h = f.read()

        assert eq(len(a), 10)
        assert eq(len(b), 10)

        assert eq(len(e), 101 - 10)
        assert eq(len(f), 101 - 10)

    xyz = NetCDFTrajectoryFile(get_fn('mdcrd.nc')).read()[0]

    assert eq(a, xyz[0:10])
    assert eq(e, xyz[10:])


def test_read_chunk_2(get_fn):
    with NetCDFTrajectoryFile(get_fn('mdcrd.nc')) as f:
        a, b, c, d = f.read(10)
        e, f, g, h = f.read(100000000000)

        assert eq(len(a), 10)
        assert eq(len(b), 10)

        assert eq(len(e), 101 - 10)
        assert eq(len(f), 101 - 10)

    xyz = NetCDFTrajectoryFile(get_fn('mdcrd.nc')).read()[0]

    assert eq(a, xyz[0:10])
    assert eq(e, xyz[10:])


def test_read_chunk_3(get_fn):
    # too big of a chunk should not be an issue
    a = NetCDFTrajectoryFile(get_fn('mdcrd.nc')).read(1000000000)
    b = NetCDFTrajectoryFile(get_fn('mdcrd.nc')).read()

    eq(a[0], b[0])


def test_read_write_1():
    xyz = np.random.randn(100, 3, 3)
    time = np.random.randn(100)
    boxlengths = np.random.randn(100, 3)
    boxangles = np.random.randn(100, 3)

    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(xyz, time, boxlengths, boxangles)

    with NetCDFTrajectoryFile(temp) as f:
        a, b, c, d = f.read()
        assert eq(a, xyz)
        assert eq(b, time)
        assert eq(c, boxlengths)
        assert eq(d, boxangles)


def test_read_write_2(get_fn):
    xyz = np.random.randn(5, 22, 3)
    time = np.random.randn(5)

    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(xyz, time)

    with NetCDFTrajectoryFile(temp) as f:
        rcoord, rtime, rlengths, rangles = f.read()
        assert eq(rcoord, xyz)
        assert eq(rtime, time)
        assert eq(rlengths, None)
        assert eq(rangles, None)

    t = md.load(temp, top=get_fn('native.pdb'))
    eq(t.unitcell_angles, None)
    eq(t.unitcell_lengths, None)


def test_ragged_1():
    # try first writing no cell angles/lengths, and then adding some
    xyz = np.random.randn(100, 3, 3)
    time = np.random.randn(100)
    cell_lengths = np.random.randn(100, 3)
    cell_angles = np.random.randn(100, 3)

    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(xyz, time)
        with pytest.raises(ValueError):
            f.write(xyz, time, cell_lengths, cell_angles)


def test_ragged_2():
    # try first writing no cell angles/lengths, and then adding some
    xyz = np.random.randn(100, 3, 3)
    time = np.random.randn(100)
    cell_lengths = np.random.randn(100, 3)
    cell_angles = np.random.randn(100, 3)

    # from mdtraj.formats import HDF5TrajectoryFile
    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(xyz, time, cell_lengths, cell_angles)
        with pytest.raises(ValueError):
            f.write(xyz, time)


def test_read_write_25():
    xyz = np.random.randn(100, 3, 3)
    time = np.random.randn(100)

    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(xyz, time)
        f.write(xyz, time)

    with NetCDFTrajectoryFile(temp) as f:
        a, b, c, d = f.read()
        assert eq(a[0:100], xyz)
        assert eq(b[0:100], time)
        assert eq(c, None)
        assert eq(d, None)

        assert eq(a[100:], xyz)
        assert eq(b[100:], time)
        assert eq(c, None)
        assert eq(d, None)


def test_write_3():
    xyz = np.random.randn(100, 3, 3)
    time = np.random.randn(100)

    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        # you can't supply cell_lengths without cell_angles
        with pytest.raises(ValueError):
            f.write(np.random.randn(100, 3, 3), cell_lengths=np.random.randn(100, 3))
        # or the other way around
        with pytest.raises(ValueError):
            f.write(np.random.randn(100, 3, 3), cell_angles=np.random.randn(100, 3))


def test_n_atoms():
    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(np.random.randn(1, 11, 3))
    with NetCDFTrajectoryFile(temp) as f:
        eq(f.n_atoms, 11)


def test_do_overwrite():
    with open(temp, 'w') as f:
        f.write('a')

    with NetCDFTrajectoryFile(temp, 'w', force_overwrite=True) as f:
        f.write(np.random.randn(10, 5, 3))


def test_do_overwrite():
    with open(temp, 'w') as f:
        f.write('a')

    with pytest.raises(IOError):
        with NetCDFTrajectoryFile(temp, 'w', force_overwrite=False) as f:
            f.write(np.random.randn(10, 5, 3))


def test_trajectory_save_load(get_fn):
    t = md.load(get_fn('native.pdb'))
    t.unitcell_lengths = 1 * np.ones((1, 3))
    t.unitcell_angles = 90 * np.ones((1, 3))

    t.save(temp)
    t2 = md.load(temp, top=t.topology)

    eq(t.xyz, t2.xyz)
    eq(t.unitcell_lengths, t2.unitcell_lengths)


@needs_cpptraj
def test_cpptraj(get_fn):
    trj0 = md.load(get_fn('frame0.dcd'), top=get_fn('frame0.pdb'))
    trj0.save(temp)

    top = get_fn("frame0.pdb")
    subprocess.check_call([
        'cpptraj',
        '-p', top,
        '-y', temp,
        '-x', temp2
    ])

    trj1 = md.load(temp, top=top)
    trj2 = md.load(temp2, top=top)

    np.testing.assert_array_almost_equal(trj0.xyz, trj2.xyz)
    np.testing.assert_array_almost_equal(trj1.xyz, trj2.xyz)
    np.testing.assert_array_almost_equal(trj0.unitcell_vectors,
                                         trj2.unitcell_vectors)
    np.testing.assert_array_almost_equal(trj1.unitcell_vectors,
                                         trj2.unitcell_vectors)

    np.testing.assert_array_almost_equal(trj0.time, trj1.time)
    np.testing.assert_array_almost_equal(trj0.time, trj2.time)
    np.testing.assert_array_almost_equal(trj1.time, trj2.time)
